/*****************************************************************************/
/* 
 * FILENAME			: a9logger.c
 *
 * DESCRIPTION		: Connects to the TX module to log telemetry
 *
 * PUBLIC FUNCTIONS : -
 *
 * NOTES 			: - 
 * 
 * AUTHOR			: Diederich Kroeske (dkroeske@d68.nl)
 *
 * VERSION INFO		: V1.0	dkroeske		21dec2011	Released version.
 */
 /*****************************************************************************/

#include <avr/interrupt.h>
#include <avr/io.h>
#include <util/twi.h>
#include <util/delay.h>

#include "usart.h"
#include "led.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>



// defines


// Prototypes
void i2c_init_sr(unsigned char slaveAddr);
char nibble2asc(uint8_t nibble);
void init_WiFly(void);
void commandDispatcher(char *cmd);
void handleReturnAllMeasurements(void);

// Global variables;
typedef struct
{
	uint8_t frameType_0x2B[25];
	uint8_t frameType_0x2D[25];
} MSG_BUFFER_STRUCT;

MSG_BUFFER_STRUCT frameBuffer[2];
MSG_BUFFER_STRUCT *pframeBufferRead = &frameBuffer[0];
MSG_BUFFER_STRUCT *pframeBufferWrite = &frameBuffer[1];

uint8_t requestNewMsg = 0;

uint16_t msCnt;

char serBuffer[20];
uint8_t serBufferIndex = 0;

char xmlbuf[50] = {'\0'};

uint8_t initWiFlyOk;
uint8_t tcpConnectionOpen;


/*****************************************************************************/
int main(void)
/*
 * Description	: Main loop.
 * input		: -
 * output		: -
 * 
 */
/*****************************************************************************/
{		
	uint8_t idx;
//	uint8_t temp;
	
	// Disable global interrupts
	cli();
	
	// Init Timer 0
	TCCR0 = 0x05; //0x08;
	TIMSK |= (1 << TOIE0);
	TCNT0 = 0xFF - 156;		// isr every 20 ms
	TIFR |= (1 << TOV0);

	// Init led and switch led off
	initLed();

	// Init USART 
	init_uart(9600, commandDispatcher);

	
	// Init the i2c link to the optima 7/9 receivers.
	i2c_init_sr(0x00);
	TWCR |= (1<<TWINT); // clear interrupt flag, allow TW interrupts
	
	// From this point on allow interrupts
	sei();


	for(idx = 0; idx < 8; idx++ )
	{
		_delay_ms(250);
	}
	// Init WiFly module. This set Adhoc network mode. After that
	// WiFly will restart. Then, try to connect till WiFly module 
	// echo's *READY*	
	initWiFlyOk = 0;
	flushUartBuffer();
	init_WiFly();
	while (0 == initWiFlyOk)
	{
		init_WiFly();
		for(idx = 0; idx < 8; idx++ )
		{
			_delay_ms(250);
		}
	}
	
	// Indicate board startup by flashing 5x led
//	for( idx = 0; idx < 10; idx++)
//	{
//		toggleLed();
//		_delay_ms(250);
//	}
//	setLedOff();
//	_delay_ms(250);

	// Enter the forever-loop
	while(1==1)
	{ 		
// 		if( !tcpConnectionOpen )
// 		{
// 			// Fast toggle led if not connected
// 			for( idx = 0; idx < 10; idx++)
// 			{
// 				toggleLed();
//  				_delay_ms(50);
//  			}
//  			setLedOff();
//  		}
//  		else
//  		{
//  			setLedOn();
//  		}

// 		Looptime approx 1 sec
 		for( idx = 0; idx < 100; idx++)
 		{
 			_delay_ms(40);
 		}
 		
// Printout buffer;
//    		txstr("frameType_0x2B: ");
//  		for( idx = 0; idx < 20; idx++ )
//  		{
//  			temp = pframeBufferWrite->frameType_0x2B[idx];
//  			tx(nibble2asc((temp>>4)&0x0F));
//  			tx(nibble2asc(temp&0x0F));
//  			tx(' ');
//  		}
//   		txstr("\n\r");
//  
//   		txstr("frameType_0x2D: ");
//  		for( idx = 0; idx < 20; idx++ )
//  		{
//  			temp = pframeBufferWrite->frameType_0x2D[idx];
//  			tx(nibble2asc((temp>>4)&0x0F));
//  			tx(nibble2asc(temp&0x0F));
//  			tx(' ');
//  		}
//  		txstr("\n\r\n\r");

	}

	return 0;
}


/*****************************************************************************/
void init_WiFly(void)
/*
 * Description	: Init the WiFly module
 * input		: -
 * output		: -
 * 
 */
{
	uint8_t idx;
	// Make sure not in command mode
	txstr("exit\r");
	_delay_ms(250);
	
	// Enter command mode
	txstr("$$$");
	for(idx = 0; idx < 4; idx++ ) _delay_ms(250);
	
	// Network settings
	txstr("set ip a 192.168.23.10\r"); 
//	for(idx = 0; idx < 1; idx++ ) _delay_ms(250);
	txstr("set ip dhcp 0\r");
	txstr("set ip netmask 255.255.255.0\r");
	txstr("set wlan join 4\r");
	txstr("set wlan ssid a9log-3\r");
	txstr("set wlan chan 1\r");
	txstr("save\r");
	_delay_ms(250); _delay_ms(250);
	
	// Reboot
	txstr("reboot\r");
	_delay_ms(200); _delay_ms(250);
}


char nibble2asc(uint8_t nibble)
{
	if( nibble < 10 )
		return nibble + '0';
	else
		return nibble + 'A' - 10;
}

char *hex2asc(char *buf, uint8_t val)
{
	buf[0] = nibble2asc((val>>4)&0x0F);
	buf[1] = nibble2asc(val&0x0F);
	buf[2] = '\0';
	return buf;
}


/*****************************************************************************/
void commandDispatcher(char *cmd)
/*
 * Description	: This is the callbackfunction called by the serial routine as
 *				  soon as a line of text is received.
 * input		: -
 * output		: -
 * 
 */
{
	uint8_t idx = 0;
	
	if( 0 == strcmp(cmd,"hey") )
	{
		txstr("alive & kicking\n\r");
	}
	
	if( 0 == strcmp(cmd,"get all") )
	{
		handleReturnAllMeasurements();
		requestNewMsg = 1;
		
		for( idx = 0; idx < 4; idx++)
 		{
 			toggleLed();
  			_delay_ms(50);
  		}
	}
	
	if( 0 == strcmp(cmd,"*READY*") )
	{
		initWiFlyOk = 1;
	}
	
	if( 0 == strcmp(cmd,"*OPEN*") )
	{
		tcpConnectionOpen = 1;
		setLedOn();
	}

	if( 0 == strcmp(cmd,"*CLOS*") )
	{
		tcpConnectionOpen = 0;
		setLedOff();
	}
}

void handleReturnAllMeasurements(void)
{
	char tmp[8] = {'\0'};

	// Start with empty string
	xmlbuf[0] = '\0';

	strcat(xmlbuf,"<a9>");
	// Receiver voltage
	strcat(xmlbuf,"<rxvolthigh>");
	strcat(xmlbuf, hex2asc(tmp,pframeBufferWrite->frameType_0x2D[3]) );
	strcat(xmlbuf,"</rxvolthigh>");
	strcat(xmlbuf,"<rxvoltlow>");
	strcat(xmlbuf, hex2asc(tmp,pframeBufferWrite->frameType_0x2D[4]) );
	strcat(xmlbuf,"</rxvoltlow>");
	strcat(xmlbuf,"</a9>");

	strcat(xmlbuf,"\n\r");

	txstr(xmlbuf);
}


/*****************************************************************************/
ISR(TIMER0_OVF_vect)
/*
 * Description	: Timer 0 overflow ISR. Interrupt every 20ms @ 8 MHz
 * input		: -
 * output		: -
 * 
 */
{
	msCnt++;
	
	TCNT0 = 0xFF - 156;
}

/*****************************************************************************/
void i2c_init_sr(unsigned char slaveAddr)
/*
 * Description	: Initialise I2C bus (TWI in Atmel terms) for SLA+R mode.
 *
 *				  Only listen to 'slaveAddr' messages, does not respond to
 *				  general call address.
 *
 * input		: slaveAddr is I2C slave address of Hitec sensorstations
 * output		: - (ISR)
 */
{
	TWAR = slaveAddr; 

	TWDR = 0x00;
    
    TWCR = 0x00;
	TWCR = 	(1<<TWEN) |	// Enable TWI
			(1<<TWEA) | // Enable ACK
			(1<<TWIE) | // enable interrupts;
			(0<<TWSTA) | (0<<TWSTO);
}


/*****************************************************************************/
ISR(TWI_vect, ISR_BLOCK)
/*
 * Description	: I2C slave receiver ISR
 *
 * input		: - (ISR)
 * output		: - (ISR)
 */
{
	static uint8_t msgIndex = 0;
	static uint8_t frameSelector;
	static uint8_t rejectFrame = 1;
	MSG_BUFFER_STRUCT *temp;
	uint8_t data;

	uint8_t twiStatus = TWSR & 0xF8;
	
	switch(twiStatus)
	{
		// SLA+W has been received, return ACK;
		case 0x60:
		case 0x70:
			frameSelector = 0x00;
			rejectFrame = 1;
			msgIndex = 0;
			TWCR =  (1<<TWEN) | 
					(1<<TWIE) | 
					(1<<TWINT)|
					(1<<TWEA) |
					(0<<TWSTA)|
					(0<<TWSTO)|
					(0<<TWWC); 
			break;
		
		// Previously addressed or general call address. data received
		case 0x80:
		case 0x90:
			data = TWDR;

			// Telemetry datapackets always start with 0xAF otherwise reject msg
			if( 0 == msgIndex && 0xAF == data )
			{
				rejectFrame = 0;
			}

			if( !rejectFrame)
			{
				if( msgIndex == 2 )
				{
					frameSelector = data;
				}

				switch(frameSelector)
				{
					case 0x00:
						pframeBufferWrite->frameType_0x2B[msgIndex] = data;
						pframeBufferWrite->frameType_0x2D[msgIndex] = data;
						break;
		
					case 0x2B:
						pframeBufferWrite->frameType_0x2B[msgIndex] = data;
						break;
		
					case 0x2D:
						pframeBufferWrite->frameType_0x2D[msgIndex] = data;
						break;
		
					default:
						break;
				}
			}
			msgIndex++;

			// 
			TWCR =  (1<<TWEN) | 
					(1<<TWIE) | 
					(1<<TWINT)|
					(1<<TWEA) |
					(0<<TWSTA)|
					(0<<TWSTO)|
					(0<<TWWC);

			break;			
		
		case 0xC0:			
		case 0x88:
		case 0xC8:
		case 0xA0:
		default:
			// Copy save databuffer and reset TWI interface
			if( requestNewMsg )
			{
				temp = pframeBufferWrite;
				pframeBufferWrite = pframeBufferRead;
				pframeBufferRead = temp;
				requestNewMsg = 0;
			}
			TWCR = (1<<TWEN)|(1<<TWIE)|(1<<TWINT)|(1<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|(0<<TWWC);
			break;
	}
}

